# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""CIFAR10 small images classification dataset.
"""

import sys as _sys

from keras.datasets.cifar10 import load_data